#.rst:
# CTestCoverageCollectJavaJSON
# ------------------------
#
# This module provides the function ``ctest_coverage_collect_java_json``.
# The function will package all of the .java.json files it finds into a
# tar file with a subprojects.json that indicates the correspondence
# between Java packages and CDash subprojects.  This file can be sent
# to a CDash server for display with the
# :command:`ctest_submit(CDASH_UPLOAD)` command.
#
# .. command:: cdash_coverage_collect_java_json
#
#   ::
#
#     ctest_coverage_collect_java_json(TARBALL <tarfile>
#       [BUILD <build_dir>]
#       )
#
#   Package a tar file of .java.json files for CDash.  The options are:
#
#   ``TARBALL <tarfile>``
#     Specify the location of the ``.tar`` file to be created for later
#     upload to CDash.  Relative paths will be interpreted with respect
#     to the top-level build directory.
#
#   ``BUILD <build_dir>``
#     Specify the top-level build directory for the build.
#     Default is the value of :variable:`CTEST_BINARY_DIRECTORY`.

include(CMakeParseArguments)
function(ctest_coverage_collect_java_json)
  set(options "")
  set(oneValueArgs TARBALL BUILD)
  cmake_parse_arguments(JavaJSON  "${options}" "${oneValueArgs}"
    "${multiValueArgs}" "" ${ARGN} )
  if(NOT DEFINED JavaJSON_TARBALL)
    message(FATAL_ERROR
      "TARBALL must be specified. for ctest_coverage_collect_java_json")
  endif()
  if(NOT DEFINED JavaJSON_BUILD)
    set(binary_dir "${CTEST_BINARY_DIRECTORY}")
  else()
    set(binary_dir "${JavaJSON_BUILD}")
  endif()

  # look for coverage files in <bin>/coverage.
  set(json_dir "${binary_dir}/coverage")
  file(GLOB_RECURSE cov_files RELATIVE ${json_dir}
    "${json_dir}/*.java.json")

  # return early if no coverage files were found
  list(LENGTH cov_files len)
  if(len EQUAL 0)
    message("ctest_coverage_collect_java_json: No .java.json files found, "
      "ignoring coverage request.")
    return()
  endif()

  # setup the dir for the coverage files
  set(coverage_dir "${binary_dir}/Testing/CoverageInfo")
  file(MAKE_DIRECTORY  "${coverage_dir}")

  # copy each file into our coverage dir.
  foreach (cov_file ${cov_files})
    file(COPY ${json_dir}/${cov_file} DESTINATION ${coverage_dir})
  endforeach()

  # create the tarball
  execute_process(COMMAND
    ${CMAKE_COMMAND} -E tar cvfj ${JavaJSON_TARBALL}
    "--mtime=1970-01-01 0:0:0 UTC"
    "Testing/CoverageInfo"
    WORKING_DIRECTORY ${binary_dir})

endfunction()
